#ifndef REDUCEDMODELOPTIMIZER_HPP
#define REDUCEDMODELOPTIMIZER_HPP

#include "beamformfinder.hpp"
#include "edgemesh.hpp"
#include "elementalmesh.hpp"
#include "matplot/matplot.h"
#include <Eigen/Dense>

using FullPatternVertexIndex = VertexIndex;
using ReducedPatternVertexIndex = VertexIndex;

class ReducedModelOptimizer {
  std::shared_ptr<SimulationMesh> m_pReducedPatternSimulationMesh;
  std::shared_ptr<SimulationMesh> m_pFullModelSimulationMesh;
  std::unordered_map<FullPatternVertexIndex, ReducedPatternVertexIndex>
      m_fullToReducedInterfaceViMap;
  std::unordered_map<FullPatternVertexIndex, FullPatternVertexIndex>
      m_fullPatternOppositeInterfaceViMap;
  std::unordered_map<size_t, size_t> nodeToSlot;
  std::unordered_map<size_t, std::unordered_set<size_t>> slotToNode;
  std::vector<double> initialGuess;

public:
  enum SimulationScenario {
    Axial,
    Shear,
    Bending,
    Double,
    Saddle,
    NumberOfSimulationScenarios
  };
  inline static const std::string SimulationScenarioStrings[] = {
      "Axial", "Shear", "Bending", "Double", "Saddle"};
  void optimize(const int &simulationScenario = -1);
  double operator()(const Eigen::VectorXd &x, Eigen::VectorXd &) const;

  ReducedModelOptimizer(const std::vector<size_t> &numberOfNodesPerSlot);
  void computeReducedModelSimulationJob(
      const SimulationJob &simulationJobOfFullModel,
      SimulationJob &simulationJobOfReducedModel);

  SimulationJob
  getReducedSimulationJob(const SimulationJob &fullModelSimulationJob);

  void initializePatterns(
      FlatPattern &fullPattern, FlatPattern &reducedPatterm,
      const std::unordered_set<size_t> &reducedModelExcludedEges);

  void setInitialGuess(std::vector<double> v);

private:
  void computeDesiredReducedModelDisplacements(
      const SimulationResults &fullModelResults,
      Eigen::MatrixX3d &optimalDisplacementsOfReducedModel);
  Eigen::VectorXd runOptimization(double (*pObjectiveFunction)(long,
                                                               const double *));
  std::vector<SimulationJob>
  createScenarios(const std::shared_ptr<SimulationMesh> &pMesh);
  void computeMaps(FlatPattern &fullModel, FlatPattern &reducedPattern,
                   const std::unordered_set<size_t> &reducedModelExcludedEges);
  void createSimulationMeshes(FlatPattern &fullModel,
                              FlatPattern &reducedModel);
  void initializeStiffnesses();

  static double
  computeError(const SimulationResults &reducedPatternResults,
               const Eigen::MatrixX3d &optimalReducedPatternDisplacements);
  static double objectiveSingleScenario(long n, const double *x);
  static double objectiveAllScenarios(long n, const double *x);
};

#endif // REDUCEDMODELOPTIMIZER_HPP
